"""
seg_silero.py

segment a dataset with the Silero segmenter

"""

import torch
import pandas as pd
from tqdm import tqdm
import audformat
from audformat.utils import to_filewise_index
from audformat import segmented_index
from nkululeko.utils.util import Util

# from nkululeko.constants import SAMPLING_RATE

SAMPLING_RATE = 16000

vad_model, vad_utils = torch.hub.load(
    repo_or_dir="snakers4/silero-vad",
    model="silero_vad",
    force_reload=False,
    onnx=False,
)


class Silero_segmenter:
    def __init__(self, not_testing=True):
        # initialize the VAD model
        torch.set_num_threads(1)
        self.no_testing = not_testing
        self.util = Util(has_config=not_testing)

    def get_segmentation(self, file):
        #    print(f'segmenting {file[0]}')
        (
            get_speech_timestamps,
            save_audio,
            read_audio,
            VADIterator,
            collect_chunks,
        ) = vad_utils
        SAMPLING_RATE = 16000
        if self.no_testing:
            min_length = float(self.util.config_val("SEGMENT", "min_length", 2))
            max_length = float(self.util.config_val("SEGMENT", "max_length", 10))
        else:
            min_length = 2
            max_length = 10
        wav = read_audio(file[0], sampling_rate=SAMPLING_RATE)
        speech_timestamps = get_speech_timestamps(
            wav, vad_model, sampling_rate=SAMPLING_RATE
        )
        files, starts, ends = [], [], []
        for entry in speech_timestamps:
            start = float(entry["start"] / SAMPLING_RATE)
            end = float(entry["end"] / SAMPLING_RATE)
            new_end = end
            handled = False
            while end - start > max_length:
                new_end = start + max_length
                if end - new_end < min_length:
                    new_end = end
                files.append(file[0])
                starts.append(start)
                ends.append(new_end)
                start += max_length
                handled = True
            if not handled and end - start > min_length:
                files.append(file[0])
                starts.append(start)
                ends.append(end)
        seg_index = segmented_index(files, starts, ends)
        return seg_index

    def segment_dataframe(self, df):
        dfs = []
        for file, values in tqdm(df.iterrows()):
            index = self.get_segmentation(file)
            dfs.append(
                pd.DataFrame(
                    values.to_dict(),
                    index,
                )
            )
        return audformat.utils.concat(dfs)


def main():
    #    files = pd.Series(['test_wavs/47_CF61_2_7.wav'])
    files = pd.Series(["test_wavs/very_long.wav"])
    df_sample = pd.DataFrame(index=files)
    df_sample["target"] = "anger"
    df_sample.index = audformat.utils.to_segmented_index(
        df_sample.index, allow_nat=False
    )
    df_sample.head(10)
    segmenter = Silero_segmenter(not_testing=False)
    df_seg = segmenter.segment_dataframe(df_sample)

    def calc_dur(x):
        starts = x[1]
        ends = x[2]
        return (ends - starts).total_seconds()

    df_seg["duration"] = df_seg.index.to_series().map(lambda x: calc_dur(x))
    print(df_seg.head(100))


if __name__ == "__main__":
    main()



# xgbmodel.py

from xgboost import XGBClassifier
from nkululeko.models.model import Model


class XGB_model(Model):
    """An XGBoost model"""

    is_classifier = True

    clf = XGBClassifier()  # set up the classifier



# feats_praat.py
from nkululeko.feat_extract.featureset import Featureset
import os
import pandas as pd
import numpy as np
import nkululeko.glob_conf as glob_conf
from nkululeko.feat_extract import feinberg_praat
import ast


class Praatset(Featureset):
    """
    a feature extractor for the Praat software, based on
    David R. Feinberg's Praat scripts for the parselmouth python interface.
    https://osf.io/6dwr3/

    """

    def __init__(self, name, data_df):
        super().__init__(name, data_df)

    def extract(self):
        """Extract the features based on the initialized dataset or re-open them when found on disk."""
        store = self.util.get_path("store")
        store_format = self.util.config_val("FEATS", "store_format", "pkl")
        storage = f"{store}{self.name}.{store_format}"
        extract = self.util.config_val("FEATS", "needs_feature_extraction", False)
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            self.util.debug("extracting Praat features, this might take a while...")
            self.df = feinberg_praat.compute_features(self.data_df.index)
            self.df = self.df.set_index(self.data_df.index)
            for i, col in enumerate(self.df.columns):
                if self.df[col].isnull().values.any():
                    self.util.debug(
                        f"{col} includes {self.df[col].isnull().sum()} nan,"
                        " inserting mean values"
                    )
                    self.df[col] = self.df[col].fillna(self.df[col].mean())

            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug(f"reusing extracted Praat features: {storage}.")
            self.df = self.util.get_store(storage, store_format)
        self.util.debug(f"praat feature names: {self.df.columns}")
        self.df = self.df.astype(float)

    def extract_sample(self, signal, sr):
        import audiofile, audformat

        tmp_audio_names = ["praat_audio_tmp.wav"]
        audiofile.write(tmp_audio_names[0], signal, sr)
        df = pd.DataFrame(index=tmp_audio_names)
        index = audformat.utils.to_segmented_index(df.index, allow_nat=False)
        df = feinberg_praat.compute_features(index)
        df.set_index(index)
        for i, col in enumerate(df.columns):
            if df[col].isnull().values.any():
                self.util.debug(
                    f"{col} includes {df[col].isnull().sum()} nan,"
                    " inserting mean values"
                )
                mean_val = df[col].mean()
                if not np.isnan(mean_val):
                    df[col] = df[col].fillna(mean_val)
                else:
                    df[col] = df[col].fillna(0)
        df = df.astype(float)
        feats = df.to_numpy()
        return feats



# model_tree_reg.py

from sklearn.tree import DecisionTreeRegressor
from nkululeko.models.model import Model


class Tree_reg_model(Model):
    is_classifier = False

    """An Tree model"""

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.clf = DecisionTreeRegressor()  # set up the classifier



# explore.py
# explore the feature sets

from nkululeko.experiment import Experiment
import configparser
from nkululeko.utils.util import Util
from nkululeko.constants import VERSION
import argparse
import os


def main(src_dir):
    parser = argparse.ArgumentParser(description="Call the nkululeko framework.")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    util = Util("explore")
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}")

    plot_feats = eval(util.config_val("EXPL", "feature_distributions", "False"))
    tsne = eval(util.config_val("EXPL", "tsne", "False"))
    scatter = eval(util.config_val("EXPL", "scatter", "False"))
    spotlight = eval(util.config_val("EXPL", "spotlight", "False"))
    model_type = util.config_val("EXPL", "model", False)
    plot_tree = eval(util.config_val("EXPL", "plot_tree", "False"))
    needs_feats = False
    if plot_feats or tsne or scatter or model_type or plot_tree:
        # these investigations need features to explore
        expr.extract_feats()
        needs_feats = True
    # explore
    expr.analyse_features(needs_feats)
    expr.store_report()
    print("DONE")


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # use this if you want to state the config file path on command line



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# copied from librosa.util.files.py

"""Utility functions for dealing with files"""
from __future__ import annotations

import glob
import os
from pathlib import Path
from typing import Any, List, Optional, Set, Union

# add new function here
__all__ = [
    "find_files",
]


def find_files(
    directory: Union[str, os.PathLike[Any]],
    *,
    ext: Optional[Union[str, List[str]]] = None,
    recurse: bool = True,
    case_sensitive: bool = False,
    relative: bool = False,
    path_object: bool = False,
    limit: Optional[int] = None,
    offset: int = 0,
) -> List[str]:
    """Get a sorted list of (audio) files in a directory or directory sub-tree.

    Examples
    --------
    >>> from nkululeko.utils import files
    >>> # Get all audio files in a directory sub-tree
    >>> files = files.find_files('~/Music')

    >>> # Look only within a specific directory, not the sub-tree
    >>> files = files.find_files('~/Music', recurse=False)

    >>> # Only look for mp3 files and return a list of pathlib.Path objects
    >>> files = files.find_files('~/Music', ext='mp3', path_object=True)

    >>> # Or just mp3 and ogg
    >>> files = files.find_files('~/Music', ext=['mp3', 'ogg'])

    >>> # Only get the first 10 files and relative paths
    >>> files = files.find_files('~/Music', limit=10, relative=True)

    >>> # Or last 10 files
    >>> files = files.find_files('~/Music', offset=-10)

    >>> # Avoid including search patterns in the path string
    >>> import glob
    >>> directory = '~/[202206] Music'
    >>> directory = glob.escape(directory)  # Escape the special characters
    >>> files = files.find_files(directory)

    Parameters
    ----------
    directory : str
        Path to look for files

    ext : str or list of str
        A file extension or list of file extensions to include in the search.

        Default: ``['aac', 'au', 'flac', 'm4a', 'mp3', 'ogg', 'wav']``

    recurse : boolean
        If ``True``, then all subfolders of ``directory`` will be searched.

        Otherwise, only ``directory`` will be searched.

    case_sensitive : boolean
        If ``False``, files matching upper-case version of
        extensions will be included.

    path_object : boolean
        If ``True``, then return a list of ``pathlib.Path`` objects.
        Otherwise, return a list of strings. Default: ``False``

    limit : int > 0 or None
        Return at most ``limit`` files. If ``None``, all files are returned.

    offset : int
        Return files starting at ``offset`` within the list.

        Use negative values to offset from the end of the list.

    Returns
    -------
    files : list of str
        The list of audio files.
    """
    if ext is None:
        ext = ["aac", "au", "flac", "m4a", "mp3", "ogg", "wav"]

    elif isinstance(ext, str):
        ext = [ext]

    # Cast into a set
    ext = set(ext)

    # Generate upper-case versions
    if not case_sensitive:
        # Force to lower-case
        ext = {e.lower() for e in ext}
        # Add in upper-case versions
        ext |= {e.upper() for e in ext}

    fileset = set()

    if recurse:
        for walk in os.walk(directory):  # type: ignore
            fileset |= __get_files(walk[0], ext)
    else:
        fileset = __get_files(directory, ext)

    files = list(fileset)
    files.sort()
    files = files[offset:]
    if limit is not None:
        files = files[:limit]

    if relative:
        files = [os.path.relpath(f) for f in files]

    if path_object:
        files = [Path(f) for f in files]

    return files


def __get_files(dir_name: Union[str, os.PathLike[Any]], extensions: Set[str]):
    """Get a list of files in a single directory"""
    # Expand out the directory
    dir_name = os.path.abspath(os.path.expanduser(dir_name))

    myfiles = set()

    for sub_ext in extensions:
        globstr = os.path.join(dir_name, "*" + os.path.extsep + sub_ext)
        myfiles |= set(glob.glob(globstr))

    return myfiles



import os

src_path = 'demos/multiple_exeriments/'

classifiers = [
    {'--model': 'mlp',
    '--layers': '\"{\'l1\':64,\'l2\':16}\"'},
    {'--model': 'mlp',
    '--layers': '\"{\'l1\':64,\'l2\':16}\"',
    '--learning_rate': '.1',},
    {'--model': 'mlp',
    '--layers': '\"{\'l1\':64,\'l2\':16}\"',
    '--learning_rate': '.0001',
    '--drop': '.3',
    },
    {'--model': 'xgb',
    '--epochs':1},
    {'--model': 'svm',
    '--epochs':1},
]

features = [
    {'--feat': 'os'},
    # {'--feat': 'os', 
    # '--set': 'ComParE_2016',
    # },
    # {'--feat': 'mld'},
    # {'--feat': 'mld',
    # '--with_os': 'True',
    # },
    # {'--feat': 'xbow'},
    # {'--feat': 'xbow',
    # '--with_os': 'True',
    # },
    # {'--feat': 'trill'},
    # {'--feat': 'wav2vec'},
]


for c in classifiers:
    for f in features:
        cmd = f'python {src_path}parse_nkulu.py '
        for item in c:
            cmd += f'{item} {c[item]} '
        for item in f:
            cmd += f'{item} {f[item]} '
        print(cmd)
        os.system(cmd)


import argparse
from pathlib import Path

import pandas as pd

emotion_map = {
    "a": "anger",
    "d": "disgust",
    "h": "happiness",
    "f": "fear",
    "s": "sadness",
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="AESDD", help="Path of AESDD directory")
    parser.add_argument("--out_dir", type=str, default=".")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    out_dir = Path(args.out_dir)

    paths = list(data_dir.glob("**/*.wav"))
    files = [file for file in paths  if file.stem != "s05 (3)"]
    names = [file.stem for file in files]
    emotion = [emotion_map[file.stem[0]] for file in files]
    speaker = [str(int(x[x.find("(") + 1: x.find(")")])) for x in names]
    gender = [["female", "male"][int(v) % 2] for v in speaker]
    language =["greek" for file in files]

    # convert to df
    df = pd.DataFrame({"file": files, "speaker": speaker, "emotion": emotion, "gender": gender})

    # print distribution per emotion
    # print(df.groupby("emotion").count()['file'])

    # allocate speaker 5 for test set
    train_df = df[df["speaker"] != "5"]
    test_df = df.drop(train_df.index)

    # save to CSV
    df.to_csv(out_dir / "aesdd.csv", index=False)
    train_df.to_csv(out_dir / "aesdd_train.csv", index=False)
    test_df.to_csv(out_dir / "aesdd_test.csv", index=False)

    print(f"Total: {len(df)}, Train: {len(train_df)}, Test: {len(test_df)}")
    
if __name__ == "__main__":
    main()



# xgrmodel.py

from xgboost.sklearn import XGBRegressor
from nkululeko.models.model import Model


class XGR_model(Model):
    """An XGBoost model"""

    is_classifier = False

    clf = XGBRegressor()  # set up the regressor



"""
feature_extractor.py

Helper class to encapsulate feature extraction methods

"""
import pandas as pd

from nkululeko.utils.util import Util


class FeatureExtractor:
    """
    Extract acoustic features from audio samples, using several feature extractors (appends the features column-wise)
    Args:
        data_df (pandas.DataFrame): dataframe with audiofile paths as index
        feats_types (array of strings): designations of acoustic feature extractors to be used
        data_name (string): names of databases that are extracted (for the caching)
        feats_designation (string): the type of split (train/test), also is used for the cache name.
    Returns:
        df (pandas.DataFrame): dataframe with same index as data_df and acoustic features in columns
    """

    # pandas dataframe to store the features (and indexed with the data from the sets)
    df = None
    data_df = None  # dataframe to get audio paths

    # def __init__
    def __init__(self, data_df, feats_types, data_name, feats_designation):
        self.data_df = data_df
        self.data_name = data_name
        self.feats_types = feats_types
        self.util = Util("feature_extractor")
        self.feats_designation = feats_designation

    def extract(self):
        # feats_types = self.util.config_val_list('FEATS', 'type', ['os'])
        self.featExtractor = None
        self.feats = pd.DataFrame()
        _scale = True
        for feats_type in self.feats_types:
            store_name = f"{self.data_name}_{feats_type}"
            if feats_type == "os":
                from nkululeko.feat_extract.feats_opensmile import Opensmileset

                self.featExtractor = Opensmileset(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "spectra":
                from nkululeko.feat_extract.feats_spectra import Spectraloader

                self.featExtractor = Spectraloader(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "trill":
                from nkululeko.feat_extract.feats_trill import TRILLset

                self.featExtractor = TRILLset(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type.startswith("wav2vec"):
                from nkululeko.feat_extract.feats_wav2vec2 import Wav2vec2

                self.featExtractor = Wav2vec2(
                    f"{store_name}_{self.feats_designation}",
                    self.data_df,
                    feats_type,
                )
            elif feats_type.startswith("hubert"):
                from nkululeko.feat_extract.feats_hubert import Hubert

                self.featExtractor = Hubert(
                    f"{store_name}_{self.feats_designation}",
                    self.data_df,
                    feats_type,
                )

            elif feats_type.startswith("wavlm"):
                from nkululeko.feat_extract.feats_wavlm import Wavlm

                self.featExtractor = Wavlm(
                    f"{store_name}_{self.feats_designation}",
                    self.data_df,
                    feats_type,
                )

            elif feats_type.startswith("spkrec"):
                from nkululeko.feat_extract.feats_spkrec import Spkrec

                self.featExtractor = Spkrec(
                    f"{store_name}_{self.feats_designation}",
                    self.data_df,
                    feats_type,
                )
            elif feats_type == "audmodel":
                from nkululeko.feat_extract.feats_audmodel import AudModelSet

                self.featExtractor = AudModelSet(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "auddim":
                from nkululeko.feat_extract.feats_audmodel_dim import (
                    AudModelDimSet,
                )

                self.featExtractor = AudModelDimSet(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "agender":
                from nkululeko.feat_extract.feats_agender import (
                    AudModelAgenderSet,
                )

                self.featExtractor = AudModelAgenderSet(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "agender_agender":
                from nkululeko.feat_extract.feats_agender_agender import (
                    AgenderAgenderSet,
                )

                self.featExtractor = AgenderAgenderSet(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "snr":
                from nkululeko.feat_extract.feats_snr import SNRSet

                self.featExtractor = SNRSet(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "mos":
                from nkululeko.feat_extract.feats_mos import MOSSet

                self.featExtractor = MOSSet(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "squim":
                from nkululeko.feat_extract.feats_squim import SQUIMSet

                self.featExtractor = SQUIMSet(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "clap":
                from nkululeko.feat_extract.feats_clap import Clap

                self.featExtractor = Clap(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "praat":
                from nkululeko.feat_extract.feats_praat import Praatset

                self.featExtractor = Praatset(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "mld":
                from nkululeko.feat_extract.feats_mld import MLD_set

                self.featExtractor = MLD_set(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            elif feats_type == "import":
                from nkululeko.feat_extract.feats_import import Importset

                self.featExtractor = Importset(
                    f"{store_name}_{self.feats_designation}", self.data_df
                )
            else:
                self.util.error(f"unknown feats_type: {feats_type}")

            self.featExtractor.extract()
            self.featExtractor.filter()
            # remove samples that were not extracted by MLD
            # self.df_test = self.df_test.loc[self.df_test.index.intersection(featExtractor_test.df.index)]
            # self.df_train = self.df_train.loc[self.df_train.index.intersection(featExtractor_train.df.index)]
            self.util.debug(f"{feats_type}: shape : {self.featExtractor.df.shape}")
            self.feats = pd.concat([self.feats, self.featExtractor.df], axis=1)
        return self.feats

    def extract_sample(self, signal, sr):
        return self.featExtractor.extract_sample(signal, sr)



# process_database.py -> EmoFilm database

import os
import pandas as pd
from sklearn.model_selection import train_test_split
import numpy as np
import argparse

# load data

parser = argparse.ArgumentParser()
parser.add_argument('--data', type=str, default="./EmoFilm", 
                    help='path to data')
args = parser.parse_args()

data_excel = os.path.join(args.data, 'f_m_corpus_it_es_en_legend.xlsx')

label_female_df = pd.read_excel(data_excel,sheet_name="female")
label_male_df = pd.read_excel(data_excel,sheet_name="male")

# merge female and male df
label_df = pd.merge(label_female_df, label_male_df, how='outer')

# Use 'speaker' as the grouping object for cross-validation
groups = label_df["speaker"]

# get language from the file name
label_df['gender'] = label_df['file'].str.split('_').str[0]
label_df['language'] = label_df['file'].str[-2:]
label_df['emotion'] = label_df['file'].str[2:5]

# replace emotion labels with dictionary
emo_dict = {'ans': 'fea', 'dis': 'con', 'gio': 'hap', 'rab': 'ang', 'tri': 'sad'}
label_df['emotion'] = label_df['emotion'].map(emo_dict)

# change file to include full path
label_df['file'] = label_df['file'].apply(lambda x: f"{args.data}/wav_corpus_16k/{x}_16k.wav")

# split data into train, dev, test
# train: 80%, dev: 10%, test: 10%
train_files, test_files = train_test_split(label_df, test_size=0.2, random_state=42, stratify=label_df['emotion'])

# allocate 10% of train to dev
train_files, dev_files = train_test_split(train_files, test_size=0.125, random_state=42, stratify=train_files['emotion'])

# write to CSV
train_files.to_csv(f"emofilm_train.csv", index=False)
dev_files.to_csv(f"emofilm_dev.csv", index=False)
test_files.to_csv(f"emofilm_test.csv", index=False)




# main.py
# Demonstration code to use the ML-experiment framework

import argparse
import ast
import seaborn as sn
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np
import os
import audeer
from nkululeko.experiment import Experiment
import configparser
from nkululeko.utils.util import Util
from nkululeko.nkululeko import doit as nkulu
from nkululeko.aug_train import doit as aug_train
import nkululeko.glob_conf as glob_conf


def main(src_dir):
    parser = argparse.ArgumentParser(description="Call the nkululeko framework.")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    config = configparser.ConfigParser()
    config.read(config_file)
    datasets = config["EXP"]["databases"]
    datasets = ast.literal_eval(datasets)
    dim = len(datasets)
    results = np.zeros(dim * dim).reshape([dim, dim])
    last_epochs = np.zeros(dim * dim).reshape([dim, dim])
    # check if some data should be added to training
    try:
        extra_train = config["CROSSDB"]["train_extra"]
    except KeyError:
        extra_train = False

    for i in range(dim):
        for j in range(dim):
            # initialize config
            config = None
            config = configparser.ConfigParser()
            config.read(config_file)
            if i == j:
                dataset = datasets[i]
                print(f"running {dataset}")
                if extra_train:
                    config["DATA"]["databases"] = f"['{dataset}', '{extra_train}']"
                    config["DATA"][f"{extra_train}.split_strategy"] = "train"
                else:
                    config["DATA"]["databases"] = f"['{dataset}']"
                config["EXP"]["name"] = dataset
            else:
                train = datasets[i]
                test = datasets[j]
                print(f"running train: {train}, test: {test}")
                if extra_train:
                    config["DATA"][
                        "databases"
                    ] = f"['{train}', '{test}', '{extra_train}']"
                    config["DATA"][f"{test}.split_strategy"] = "test"
                    config["DATA"][f"{train}.split_strategy"] = "train"
                    config["DATA"][f"{extra_train}.split_strategy"] = "train"
                else:
                    config["DATA"]["databases"] = f"['{train}', '{test}']"
                    config["DATA"][f"{test}.split_strategy"] = "test"
                    config["DATA"][f"{train}.split_strategy"] = "train"
                config["EXP"]["name"] = f"{train}_vs_{test}"

            tmp_config = "tmp.ini"
            with open(tmp_config, "w") as tmp_file:
                config.write(tmp_file)
            if config.has_section("AUGMENT"):
                result, last_epoch = aug_train(tmp_config)
            else:
                result, last_epoch = nkulu(tmp_config)
            results[i, j] = float(result)
            last_epochs[i, j] = int(last_epoch)
    print(repr(results))
    print(repr(last_epochs))
    root = os.path.join(config["EXP"]["root"], "")
    plot_name = f"{root}/heatmap.png"
    plot_heatmap(results, last_epochs, datasets, plot_name, config, datasets)


def trunc_to_three(x):
    return int(x * 1000) / 1000.0


def plot_heatmap(results, last_epochs, labels, name, config, datasets):
    df_cm = pd.DataFrame(
        results, index=[i for i in labels], columns=[i for i in labels]
    )
    mean = trunc_to_three(results.mean())
    mean_diag = trunc_to_three(results.diagonal().mean())
    mean_non_diag = trunc_to_three(
        (results.sum() - results.diagonal().sum())
        / (results.size - results.diagonal().size)
    )
    colsums = results.mean(axis=0)
    vfunc = np.vectorize(trunc_to_three)
    colsums = vfunc(colsums)
    colsums_epochs = last_epochs.mean(axis=0)
    colsums_epochs = vfunc(colsums_epochs)
    res_dir = config["EXP"]["root"]
    file_name = f"{res_dir}/results.txt"
    with open(file_name, "w") as text_file:
        text_file.write(
            f"Mean UAR: {mean} (self: {mean_diag}, cross: {mean_non_diag})\n"
        )
        data_s = ", ".join(datasets)
        text_file.write(f"{data_s}\n")
        colsums = np.array2string(colsums, separator=", ")
        text_file.write(f"column sums\n{colsums}\n")
        text_file.write("all results\n")
        text_file.write(repr(results))
        text_file.write("\n")
        colsums_epochs = np.array2string(colsums_epochs, separator=", ")
        text_file.write(f"column sums epochs\n{colsums_epochs}\n")
        text_file.write("all epochs\n")
        text_file.write(repr(last_epochs))

    plt.figure(figsize=(10, 7))
    ax = sn.heatmap(df_cm, annot=True, cmap=cm.Blues)
    caption = f"Rows: train, Cols: test. Mean UAR: {mean} (self: {mean_diag}, cross: {mean_non_diag})."
    ax.set_title(caption)
    plt.savefig(name)
    plt.close()


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # sys.argv[1])



# augment.py
# augment the training sets

import argparse
import pandas as pd
import os
import ast
from nkululeko.experiment import Experiment
import configparser
from nkululeko.utils.util import Util
from nkululeko.constants import VERSION


def doit(config_file):
    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    util = Util("augment")
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")

    filename = util.config_val("AUGMENT", "result", "augmented.csv")
    filename = f"{expr.data_dir}/{filename}"

    if os.path.exists(filename):
        util.debug("files already augmented")
    else:
        # load the data
        expr.load_datasets()

        # split into train and test
        expr.fill_train_and_tests()
        util.debug(
            f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}"
        )

        # augment
        augmentings = util.config_val("AUGMENT", "augment", False)
        got_one = False
        if augmentings:
            augmentings = ast.literal_eval(augmentings)
            results = []
            if "traditional" in augmentings:
                df1 = expr.augment()
                results.append(df1)
                got_one = True
            if "random_splice" in augmentings:
                df2 = expr.random_splice()
                results.append(df2)
                got_one = True
        if not augmentings:
            util.error("no augmentation selected")
        if not got_one:
            util.error(f"invalid augmentation(s): {augmentings}")
        df_ret = pd.DataFrame()
        df_ret = pd.concat(results)
        # remove encoded labels
        target = util.config_val("DATA", "target", "emotion")
        if "class_label" in df_ret.columns:
            df_ret = df_ret.drop(columns=[target])
            df_ret = df_ret.rename(columns={"class_label": target})
        # save file
        df_ret.to_csv(filename)
        util.debug(f"saved augmentation table to {filename} to {expr.data_dir}")
        print("DONE")


def main(src_dir):
    parser = argparse.ArgumentParser(description="Call the nkululeko framework.")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"
    doit(config_file)


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # use this if you want to state the config file path on command line



""" 
model_cnn.py

Inspired by code from Su Lei

"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
import torchvision.transforms as transforms
from torch.utils.data import Dataset
import ast
import numpy as np
from sklearn.metrics import recall_score
from collections import OrderedDict
from PIL import Image

from nkululeko.utils.util import Util
import nkululeko.glob_conf as glob_conf
from nkululeko.models.model import Model
from nkululeko.reporter import Reporter
from nkululeko.losses.loss_softf1loss import SoftF1Loss


class CNN_model(Model):
    """CNN = convolutional neural net"""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor taking the configuration and all dataframes"""
        super().__init__(df_train, df_test, feats_train, feats_test)
        super().set_model_type("ann")
        self.target = glob_conf.config["DATA"]["target"]
        labels = glob_conf.labels
        self.class_num = len(labels)
        # set up loss criterion
        criterion = self.util.config_val("MODEL", "loss", "cross")
        if criterion == "cross":
            self.criterion = torch.nn.CrossEntropyLoss()
        elif criterion == "f1":
            self.criterion = SoftF1Loss(
                num_classes=self.class_num, weight=None, epsilon=1e-7
            )
        else:
            self.util.error(f"unknown loss function: {criterion}")
        self.util.debug(f"using model with cross entropy loss function")
        # set up the model
        self.device = self.util.config_val("MODEL", "device", "cpu")
        try:
            layers_string = glob_conf.config["MODEL"]["layers"]
        except KeyError as ke:
            self.util.error(f"Please provide MODEL layers: {ke}")
        self.util.debug(f"using layers {layers_string}")
        layers = ast.literal_eval(layers_string)
        # with dropout?
        drop = self.util.config_val("MODEL", "drop", False)
        if drop:
            self.util.debug(f"init: training with dropout: {drop}")
        self.model = myCNN(layers, self.class_num).to(self.device)
        self.learning_rate = float(
            self.util.config_val("MODEL", "learning_rate", 0.0001)
        )
        # set up regularization
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate
        )
        # batch size
        self.batch_size = int(self.util.config_val("MODEL", "batch_size", 8))
        # number of parallel processes
        self.num_workers = int(self.util.config_val("MODEL", "num_workers", 5))

        # set up the data_loaders

        # Define transformations
        transformations = transforms.Compose(
            [
                transforms.ToTensor()
                # transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
            ]
        )
        train_set = self.Dataset_image(
            feats_train, df_train, self.target, transformations
        )
        test_set = self.Dataset_image(feats_test, df_test, self.target, transformations)
        # Define data loaders
        self.trainloader = torch.utils.data.DataLoader(
            train_set,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
        )
        self.testloader = torch.utils.data.DataLoader(
            test_set,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
        )

    class Dataset_image(Dataset):
        def __init__(self, df_feats, df_labels, target, transform=None):
            self.df_feats = df_feats
            self.df_labels = df_labels
            self.transform = transform
            self.target = target

        def __len__(self):
            return len(self.df_feats)

        def __getitem__(self, idx):
            # Load the image file
            img_path = self.df_feats.iloc[idx, 0]
            image = Image.open(img_path)
            # Get emotion label
            label = self.df_labels[self.target].iloc[idx]
            if self.transform:
                image = self.transform(image)
            return image, label

    def set_testdata(self, data_df, feats_df):
        test_set = self.Dataset_image(feats_df, data_df)
        # Define data loaders
        self.testloader = torch.utils.data.DataLoader(
            test_set,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
        )

    def reset_test(self, df_test, feats_test):
        self.set_testdata(df_test, feats_test)

    def train(self):
        self.model.train()
        losses = []
        for images, labels in self.trainloader:
            logits = self.model(images.to(self.device))
            loss = self.criterion(logits, labels.to(self.device, dtype=torch.int64))
            losses.append(loss.item())
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
        self.loss = (np.asarray(losses)).mean()

    def evaluate_model(self, model, loader, device):
        logits = torch.zeros(len(loader.dataset), self.class_num)
        targets = torch.zeros(len(loader.dataset))
        model.eval()
        losses = []
        with torch.no_grad():
            for index, (images, labels) in enumerate(loader):
                start_index = index * loader.batch_size
                end_index = (index + 1) * loader.batch_size
                if end_index > len(loader.dataset):
                    end_index = len(loader.dataset)
                logits[start_index:end_index, :] = model(images.to(device))
                targets[start_index:end_index] = labels
                loss = self.criterion(
                    logits[start_index:end_index, :],
                    labels.to(self.device, dtype=torch.int64),
                )
                losses.append(loss.item())

        self.loss_eval = (np.asarray(losses)).mean()
        predictions = logits.argmax(dim=1)
        uar = recall_score(targets.numpy(), predictions.numpy(), average="macro")
        return uar, targets, predictions

    def predict(self):
        _, truths, predictions = self.evaluate_model(
            self.model, self.testloader, self.device
        )
        uar, _, _ = self.evaluate_model(self.model, self.trainloader, self.device)
        report = Reporter(truths, predictions, self.run, self.epoch)
        try:
            report.result.loss = self.loss
        except AttributeError:  # if the model was loaded from disk the loss is unknown
            pass
        try:
            report.result.loss_eval = self.loss_eval
        except AttributeError:  # if the model was loaded from disk the loss is unknown
            pass
        report.result.train = uar
        return report

    def get_predictions(self):
        _, truths, predictions = self.evaluate_model(
            self.model, self.testloader, self.device
        )
        return predictions.numpy()

    def predict_sample(self, features):
        """Predict one sample"""
        with torch.no_grad():
            logits = self.model(torch.from_numpy(features).to(self.device))
        a = logits.numpy()
        res = {}
        for i in range(len(a[0])):
            res[i] = a[0][i]
        return res

    def store(self):
        torch.save(self.model.state_dict(), self.store_path)

    def load(self, run, epoch):
        self.set_id(run, epoch)
        dir = self.util.get_path("model_dir")
        # name = f'{self.util.get_exp_name()}_{run}_{epoch:03d}.model'
        name = f"{self.util.get_exp_name(only_train=True)}_{self.run}_{self.epoch:03d}.model"
        self.device = self.util.config_val("MODEL", "device", "cpu")
        layers = ast.literal_eval(glob_conf.config["MODEL"]["layers"])
        self.store_path = dir + name
        drop = self.util.config_val("MODEL", "drop", False)
        if drop:
            self.util.debug(f"loading: dropout set to: {drop}")
        self.model = myCNN(layers, self.class_num).to(self.device)
        self.model.load_state_dict(torch.load(self.store_path))
        self.model.eval()

    def load_path(self, path, run, epoch):
        self.set_id(run, epoch)
        with open(path, "rb") as handle:
            self.device = self.util.config_val("MODEL", "device", "cpu")
            layers = ast.literal_eval(glob_conf.config["MODEL"]["layers"])
            self.store_path = path
            drop = self.util.config_val("MODEL", "drop", False)
            if drop:
                self.util.debug(f"dropout set to: {drop}")
            self.model = self.MLP(
                self.feats_train.shape[1], layers, self.class_num, drop
            ).to(self.device)
            self.model.load_state_dict(torch.load(self.store_path))
            self.model.eval()


class myCNN(torch.nn.Module):
    def __init__(self, layers, class_num):
        sorted_layers = sorted(layers.items(), key=lambda x: x[1])
        l1 = sorted_layers[0][1]
        l2 = sorted_layers[1][1]
        super(myCNN, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 61 * 61, l1)
        self.fc2 = nn.Linear(l1, l2)
        self.fc3 = nn.Linear(l2, class_num)

    def forward(self, x):
        # -> n, 3, 256, 256
        x = self.pool(F.relu(self.conv1(x)))  # -> n, 6, 126, 126
        x = self.pool(F.relu(self.conv2(x)))  # -> n, 16, 61, 61
        x = x.view(-1, 16 * 61 * 61)  # -> n, 59536
        x = F.relu(self.fc1(x))  # -> n, 120
        x = F.relu(self.fc2(x))  # -> n, 84
        x = self.fc3(x)  # -> n, 7
        return x



import audformat
import pandas as pd
import nkululeko.glob_conf as glob_conf
from nkululeko.utils.util import Util
import ast


class DataFilter:
    def __init__(self, df):
        self.util = Util("datafilter")
        self.df = df.copy()
        self.util.copy_flags(df, self.df)

    def all_filters(self, data_name=""):
        self.limit_samples(data_name)
        self.limit_speakers(data_name)
        self.filter_value(data_name)
        return self.filter_duration(data_name)

    def limit_samples(self, data_name=""):
        """limit number of samples
        the samples are selected randomly
        """
        if data_name == "":
            max = self.util.config_val("DATA", "limit_samples", False)
        else:
            max = self.util.config_val_data(data_name, "limit_samples", False)
        if max:
            if self.df.shape[0] < int(max):
                return self.df
            else:
                df = self.df.sample(int(max))
                self.util.debug(
                    f"{data_name}: limited samples to {max}, reduced samples"
                    f" from {self.df.shape[0]} to {df.shape[0]}"
                )
                self.util.copy_flags(self.df, df)
                self.df = df
                return df
        else:
            return self.df

    def limit_speakers(self, data_name=""):
        """limit number of samples per speaker
        the samples are selected randomly
        """
        if data_name == "":
            max = self.util.config_val("DATA", "limit_samples_per_speaker", False)
        else:
            max = self.util.config_val_data(
                data_name, "limit_samples_per_speaker", False
            )
        if max:
            df = pd.DataFrame()
            for s in self.df.speaker.unique():
                s_df = self.df[self.df["speaker"].eq(s)]
                if s_df.shape[0] < int(max):
                    df = pd.concat([df, s_df])
                else:
                    df = pd.concat([df, s_df.sample(int(max))])
            self.util.debug(
                f"{data_name}: limited samples to {max} per speaker, reduced"
                f" samples from {self.df.shape[0]} to {df.shape[0]}"
            )
            self.util.copy_flags(self.df, df)
            self.df = df
            return df
        else:
            return self.df

    def filter_duration(self, data_name=""):
        """remove all samples less than min_dur duration"""
        if data_name == "":
            min_dur = self.util.config_val("DATA", "min_duration_of_sample", False)
            max_dur = self.util.config_val("DATA", "max_duration_of_sample", False)
        else:
            min_dur = self.util.config_val_data(
                data_name, "min_duration_of_sample", False
            )
            max_dur = self.util.config_val_data(
                data_name, "max_duration_of_sample", False
            )
        if min_dur or max_dur:
            if not isinstance(self.df.index, pd.MultiIndex):
                self.util.debug(
                    "converting file index to multi index, this might take a"
                    " while..."
                )
                self.df.index = audformat.utils.to_segmented_index(
                    self.df.index, allow_nat=False
                )
            if min_dur:
                old_samples = self.df.shape[0]
                df = self.df.copy()
                for i in self.df.index:
                    start = i[1]
                    end = i[2]
                    dur = (end - start).total_seconds()
                    if dur < float(min_dur):
                        df = df.drop(i, axis=0)
                self.util.debug(
                    f"{data_name}: filtered samples less than"
                    f" {min_dur} seconds, reduced samples from {old_samples} to"
                    f" {df.shape[0]}"
                )
            if max_dur:
                old_samples = self.df.shape[0]
                df = self.df.copy()
                for i in self.df.index:
                    start = i[1]
                    end = i[2]
                    dur = (end - start).total_seconds()
                    if dur > float(max_dur):
                        df = df.drop(i, axis=0)
                self.util.debug(
                    f"{data_name}: filtered samples more than"
                    f" {max_dur} seconds, reduced samples from {old_samples} to"
                    f" {df.shape[0]}"
                )
            self.util.copy_flags(self.df, df)
            self.df = df
            return df
        else:
            return self.df

    def filter_value(self, data_name=""):
        if data_name == "":
            filters_str = self.util.config_val("DATA", "filter", False)
        else:
            filters_str = self.util.config_val_data(data_name, "filter", False)
        if filters_str:
            filters = ast.literal_eval(filters_str)
            df = self.df.copy()
            for f in filters:
                col = f[0]
                val = f[1]
                pre = df.shape[0]
                df = df[df[col] == val]
                post = df.shape[0]
                self.util.debug(
                    f"{data_name}: filtered {col}={val}, reduced samples from"
                    f" {pre} to {post}"
                )
            self.util.copy_flags(self.df, df)
            self.df = df
            return df
        else:
            return self.df


def limit_speakers(df, max=20):
    """limit number of samples per speaker
    the samples are selected randomly
    """
    df_ret = pd.DataFrame()
    for s in df.speaker.unique():
        s_df = df[df["speaker"].eq(s)]
        if s_df.shape[0] < max:
            df_ret = df_ret.append(s_df)
        else:
            df_ret = df_ret.append(s_df.sample(max))
    return df_ret


def filter_min_dur(df, min_dur):
    """remove all samples less than min_dur duration"""
    df_ret = df.copy()
    if not isinstance(df.index, pd.MultiIndex):
        glob_conf.util.debug(
            "converting file index to multi index, this might take a while..."
        )
        df_ret.index = audformat.utils.to_segmented_index(df.index, allow_nat=False)
    for i in df_ret.index:
        start = i[1]
        end = i[2]
        dur = (end - start).total_seconds()
        if dur < float(min_dur):
            df_ret = df_ret.drop(i, axis=0)
    df_ret.is_labeled = df.is_labeled
    df_ret.got_gender = df.got_gender
    df_ret.got_speaker = df.got_speaker
    return df_ret


def filter_max_dur(df, max_dur):
    """remove all samples less than min_dur duration"""
    df_ret = df.copy()
    if not isinstance(df.index, pd.MultiIndex):
        glob_conf.util.debug(
            "converting file index to multi index, this might take a while..."
        )
        df_ret.index = audformat.utils.to_segmented_index(df.index, allow_nat=False)
    for i in df_ret.index:
        start = i[1]
        end = i[2]
        dur = (end - start).total_seconds()
        if dur > float(max_dur):
            df_ret = df_ret.drop(i, axis=0)
    df_ret.is_labeled = df.is_labeled
    df_ret.got_gender = df.got_gender
    df_ret.got_speaker = df.got_speaker
    return df_ret




import sys
#sys.path.append("../../src")
sys.path.append("../nkululeko/src")
import constants
import numpy as np
import experiment as exp
import configparser
from util import Util
import argparse
import os.path
src_path = './demos/multiple_experiments/'


def main():
    parser = argparse.ArgumentParser(description='Call the nkululeko framework.')
    parser.add_argument('--config', default='exp.ini', help='The base configuration')
    parser.add_argument('--data', help='The databases', nargs='*', \
        action='append')
    parser.add_argument('--label', nargs='*', help='The labels for the target', \
        action='append')
    parser.add_argument('--tuning_params', nargs='*', help='parameters to be tuned', \
        action='append')
    parser.add_argument('--layers', nargs='*', help='layer config for mlp, e.g. l1:128 ', \
        action='append')
    parser.add_argument('--model', default='xgb', help='The model type')
    parser.add_argument('--feat', default='os', help='The feature type')
    parser.add_argument('--set', help='The opensmile set')
    parser.add_argument('--with_os', help='To add os features')
    parser.add_argument('--target', help='The target designation')
    parser.add_argument('--epochs', help='The number of epochs')
    parser.add_argument('--runs', help='The number of runs')
    parser.add_argument('--learning_rate', help='The learning rate')
    parser.add_argument('--drop', help='The dropout rate [0:1]')
    
    args = parser.parse_args()

    if args.config is not None:
        config_file = args.config    
    else:
        config_file = f'{src_path}exp.ini'
    # test if config is there
    if not os.path.isfile(config_file):
        print(f'ERROR: no such file {config_file}')

    config = configparser.ConfigParser()
    config.read(config_file)
    # fill the config


    if args.data is not None:
        databases = []
        for t in args.data:
            databases.append(t[0])
        print(f'got databases: {databases}')
        config['DATA']['databases'] = str(databases)
    if args.label is not None:
        labels = []
        for l in args.label:
            labels.append(l[0])
        print(f'got labels: {labels}')
        config['DATA']['labels'] = str(labels)
    if args.tuning_params is not None:
        tuning_params = []
        for tp in args.tuning_params:
            tuning_params.append(tp[0])
        config['MODEL']['tuning_params'] = str(tuning_params)
    if args.layers is not None:
        config['MODEL']['layers'] = args.layers[0][0]
    if args.target is not None:
        config['DATA']['target'] = args.target
    if args.epochs is not None:
        config['EXP']['epochs'] = args.epochs
    if args.runs is not None:
        config['EXP']['runs'] = args.runs
    if args.learning_rate is not None:
        config['MODEL']['learning_rate'] = args.learning_rate
    if args.drop is not None:
        config['MODEL']['drop'] = args.drop
    if args.model is not None:
        config['MODEL']['type'] = args.model
    if args.feat is not None:
        config['FEATS']['type'] = f'[\'{args.feat}\']'
    if args.with_os is not None:
        config['FEATS']['with_os'] = args.with_os
    if args.set is not None:
        config['FEATS']['set'] = args.set


    name = config['EXP']['name']

    # init the experiment
    expr = exp.Experiment(config)
    util = Util()
    util.debug(f'running {name}, Nkululeko version {constants.VERSION}')

    # load the data
    expr.load_datasets()
    # split into train and test
    expr.fill_train_and_tests()
    # extract features
    expr.extract_feats()
    # initialize a run manager
    expr.init_runmanager()
    # run the experiment
    reports = expr.run()
    result = reports[-1].result.test
    # report result
    util.debug(f'result for {expr.get_name()} is {result}')

if __name__ == "__main__":
    main()


# randomsplicer.py

"""
Code originally by Oliver Pauly

Based on an idea by Klaus Scherer

K. R. Scherer, “Randomized splicing: A note on a simple technique for masking speech content” 
Journal of Experimental Research in Personality, vol. 5, pp. 155–159, 1971.

Evaluated in:
F. Burkhardt, Anna Derington, Matthias Kahlau, Klaus Scherer, Florian Eyben and Björn Schuller: Masking Speech Contents by Random Splicing: is Emotional Expression Preserved?, Proc. ICASSP, 2023

"""

import pandas as pd
from tqdm import tqdm
import os
import numpy as np
import audiofile as af
from audformat.utils import map_file_path
import audeer
from nkululeko.utils.util import Util
import nkululeko.augmenting.randomsplicing as rsp


class Randomsplicer:
    """
    augmenting the train split
    """

    def __init__(self, df):
        self.df = df
        self.util = Util("randomsplicer")

    def run(self, sample_selection):
        """
        random splice the selected samples and return a dataframe with new files index.
            adjustable parameters:
            * p_reverse: probability of some samples to be in reverse order (default: 0.3)
            * top_db: top db level for silence to be recognized (default: 12)
        """

        p_reverse = 0.3
        top_db = 12

        files = self.df.index.get_level_values(0).values
        store = self.util.get_path("store")
        filepath = f"{store}randomspliced/"
        audeer.mkdir(filepath)
        self.util.debug(f"random splicing {sample_selection} samples to {filepath}")
        newpath = ""
        index_map = {}
        for i, f in enumerate(tqdm(files)):
            signal, sr = af.read(f)
            filename = os.path.basename(f)
            parent = os.path.dirname(f).split("/")[-1]
            sig_new = rsp.random_splicing(
                signal,
                sr,
                p_reverse=p_reverse,
                top_db=top_db,
            )
            newpath = f"{filepath}/{parent}/"
            audeer.mkdir(newpath)
            new_full_name = newpath + filename
            af.write(new_full_name, signal=sig_new, sampling_rate=sr)
            index_map[f] = new_full_name

        df_ret = self.df.copy()

        file_index = df_ret.index.to_series().map(lambda x: index_map[x[0]]).values
        # workaround because i just couldn't get this easier...
        arrays = [
            file_index,
            list(df_ret.index.get_level_values(1)),
            list(df_ret.index.get_level_values(2)),
        ]
        new_index = pd.MultiIndex.from_arrays(arrays, names=("file", "start", "end"))
        df_ret = df_ret.set_index(new_index)

        return df_ret



class DefineBase:

    @classmethod
    def _assert_has_attribute_value(cls, value):
        valid_values = cls._attribute_values()
        if value not in valid_values:
            raise BadValueError(value, valid_values)

    @classmethod
    def _attribute_values(cls):
        attributes = inspect.getmembers(
            cls, lambda x: not inspect.isroutine(x)
        )
        return sorted(
            [
                a[1] for a in attributes
                if not (a[0].startswith('__') and a[0].endswith('__'))
            ]
        )


class Header(DefineBase):
    HEADER_RESULTS = 'Results'
    HEADER_EXPLORE = 'Data exploration'



"""
This folder is to import the 
Androids-corpus depression
database to nkululeko.

I used the version downloadable from [github](https://github.com/CheyneyComputerScience/CREMA-D)

downloaded April 27th 2023

I used the version downloadable from [Dropbox, mentioned in this github page](https://github.com/androidscorpus/data)

Download and unzip the file Androids-corpus.zip to the current folder

Usage: `python process_database.py`

"""

import pandas as pd
import os
import audeer

dataset_name = 'androids'
data_root = './Androids-Corpus/'

# read in the fold list 
fold_dict = {}
df_fold = pd.read_csv(f'{data_root}fold-lists.csv')
for i in range(df_fold.shape[0]): 
    for j in range(df_fold.shape[1]): 
        try:
            value = df_fold.iloc[i, j].replace('\'', '')
        except AttributeError:
            value = 'na' 
        if j in range(7, 12):
            fold_dict[value] = j - 7
        else:
            fold_dict[value] = j 
directory_list = audeer.list_file_names(data_root, filetype='wav', recursive=True, basenames=True)

depressions, speakers, educations, genders, ages, tasks, folds = [], [], [], [], [], [], []
file_paths = []
print(len(directory_list))
gender_map = {'F':'female', 'M':'male'}
depression_map = {'P':'depressed', 'C':'control'}

for file in directory_list:
    # storing file paths
    # file = file.replace(os.getcwd(), '.')
    file_paths.append(data_root+file)
    # storing file emotions
    fn = audeer.basename_wo_ext(file)

    # The naming convention of the audio files is as follows:
    # nn_XGmm_t.wav
    # where nn is a unique integer identifier such that, in a given group, files with the same nn contain the voice of the same speaker (there is a trailing 0 for numbers lower than 10), X is an alphabetic character corresponding to the speaker’s condition (P for depression patient and C for control), G is an alphabetic character that stands for the speaker’s gender (M for male and F for female), mm is a two-digits integer number corresponding to the speaker’s age, and t is an integer number between 1 and 4 accounting for the education level (1 corresponds to primary school and 4 corresponds to university). The letter X was used for the 2 participants who did not provide information about this aspect. There is no indication of the task because recordings corresponding to RT and IT are stored in different directories.
    if 'Reading-Task'in file:
        task = 'reading'
    elif 'Interview-Task' in file:
        task = 'interview'
    else:
        print('ERROR: task undefined')
        exit(-1)
 
    part = fn.split('_')
    dir_name = f'{part[0]}_{part[1]}_{part[2]}'
    depression = part[1][0]
    speaker = f'{depression}_{part[0]}'
    gender = part[1][1]
    age = part[1][2:4]
    education = part[2]
    depressions.append(depression_map[depression])
    speakers.append(speaker)
    genders.append(gender_map[gender])
    ages.append(age)
    tasks.append(task)
    educations.append(education)
    folds.append(fold_dict[dir_name])
#    print(f'{file} {speaker}')


# dataframe for emotion of files
df = pd.DataFrame({'file':file_paths, 
                   'speaker':speakers, 
                   'gender':genders, 
                   'age':ages, 
                   'task':tasks, 
                   'depression':depressions, 
                   'education':educations, 
                   'fold':folds})

df = df.set_index('file')
df.head()
df.to_csv(f'{dataset_name}.csv')


"""
code by Anna Derington

"""
import json
import os
import shutil

import numpy as np
import pandas as pd

import audeer
import audformat


def main():
    description = (
        'Emozionalmente is an extensive, crowdsourced Italian emotional speech corpus. '
        'The dataset consists of 6902 labeled samples acted out by 431 amateur actors '
        'while verbalizing 18 different sentences expressing the Big Six emotions '
        '(anger, disgust, fear, joy, sadness, surprise) plus neutrality. '
        'Labels represent the emotional communicative intention of the actors '
        '(i.e., the seven emotional states). '
        'Recordings were generally obtained with non-professional equipment. '
        'They are .wav files, mono-channel, and have a sample size of 16 bits and a sample rate of 16000 Hz. '
        'Each audio recording lasts 3.81 seconds (SD = 0.99 seconds). '
        'The emotional content of the clips were validated by asking 829 humans (5 evaluations per audio) '
        'to guess the emotion contained in each recording. Humans obtained a general accuracy of 66%. '
    )
    languages = ['it']
    db = audformat.Database(
        name='Emozionalmente',
        source='https://zenodo.org/record/6569824#.ZGsf1hlByV4',
        usage=audformat.define.Usage.COMMERCIAL,
        description=description,
        license=audformat.define.License.CC_BY_4_0,
        languages=languages,
        author='Fabio Catania',
    )
    current_dir = os.path.dirname(__file__)
    build_dir = audeer.mkdir(os.path.join(current_dir, './build'))
    source_dir = 'emozionalmente_dataset'
    audio_dir_name = 'audio'
    audio_path = os.path.join(current_dir, source_dir, audio_dir_name)

    user_path = os.path.join(current_dir, source_dir, 'metadata', 'users.csv')
    samples_path = os.path.join(current_dir, source_dir, 'metadata', 'samples.csv')
    evaluations_path = os.path.join(current_dir, source_dir, 'metadata', 'evaluations.csv')

    emotion_labels = {
        'anger': {'original': 'anger'},
        'disgust': {'original': 'disgust'},
        'fear': {'original': 'fear'},
        'happiness': {'original':'joy'},
        'neutral': {'original': 'neutrality'},
        'no_agreement': {'original': 'no_agreement'},
        'sadness': {'original': 'sadness'},
        'surprise': {'original': 'surprise'}
    }
    emotion_mapping = {
        'anger': 'anger',
        'disgust': 'disgust',
        'fear': 'fear',
        'joy': 'happiness',
        'neutrality': 'neutral',
        'no_agreement': 'no_agreement',
        'sadness': 'sadness',
        'surprise': 'surprise'
    }

    user_df = pd.read_csv(user_path)
    user_df.set_index('username', inplace=True)
    samples_df = pd.read_csv(samples_path, 
        converters={'file_name': lambda x: os.path.join(audio_dir_name, x)})
    evaluations_df = pd.read_csv(evaluations_path,
        converters={'file_name': lambda x: os.path.join(audio_dir_name, x)})
    samples_df['emotion_expressed'] = samples_df['emotion_expressed'].map(emotion_mapping)
    evaluations_df['emotion_recognized'] = evaluations_df['emotion_recognized'].map(emotion_mapping)

    # -------- Add schemes --------
    speakers = samples_df['actor'].unique()
    speaker_info = {}
    for speaker in speakers:
        speaker_row = user_df.loc[speaker]
        speaker_info[speaker] = {
            'gender': speaker_row['gender'], 'age': int(speaker_row['age']),
            'mother_tongue': speaker_row['mother_tongue']
        }

    db.schemes['speaker'] = audformat.Scheme(
        labels=speaker_info
    )

    db.schemes['emotion'] = audformat.Scheme(
        labels=emotion_labels
    )
    db.schemes['emotion.agreement'] = audformat.Scheme(
        dtype='float', minimum=0, maximum=1,
    )

    db.schemes['votes'] = audformat.Scheme(
        dtype='int', minimum=0,
    )

    sentences = samples_df['sentence'].unique()
    transcriptid2sentence = {
        f's{i}': sentence for i, sentence in enumerate(sentences)
    }
    sentence2transcriptid = {
        v:k for k,v in transcriptid2sentence.items()
    }
    db.schemes['transcription'] = audformat.Scheme(
        labels = transcriptid2sentence
    )
    samples_df['transcription'] = samples_df['sentence'].map(sentence2transcriptid)

    # -------- Gold standard of emotion votes --------

    emotion_gold_standard = evaluations_df.groupby('file_name')['emotion_recognized'].agg(
        gold_standard
    )
    emotion_agreement = evaluations_df.groupby('file_name')['emotion_recognized'].agg(
        categorical_voter_percentage
    )
    emotion_agreement.name = 'emotion.agreement'

    samples_df = samples_df.merge(emotion_gold_standard, on='file_name')
    samples_df = samples_df.merge(emotion_agreement, on='file_name')
    samples_df.index = audformat.filewise_index(samples_df['file_name'])

    # -------- Rater values --------
    votes_df = evaluations_df.groupby(['file_name', 'emotion_recognized']).size().unstack(fill_value=0)
    votes_df.index = audformat.filewise_index(votes_df.reset_index()['file_name'])
    votes_df.reindex(samples_df.index)  # use same order for index as for other tables

    # -------- Files table with speaker information --------

    db['files'] = audformat.Table(
        index=samples_df.index
    )

    db['files']['speaker'] = audformat.Column(
        scheme_id='speaker'
    )
    db['files']['speaker'].set(samples_df['actor'])
    db['files']['transcription'] = audformat.Column(
        scheme_id='transcription'
    )
    db['files']['transcription'].set(samples_df['transcription'])


    # -------- Load speaker splits as created in the notebook --------
    speaker_split_path = os.path.join(current_dir, 'speaker_splits.json')
    with open(speaker_split_path, 'r') as fp:
        speaker_info = json.load(fp)


    for split, split_speakers in speaker_info.items():
        # Create db splits
        db.splits[split] = audformat.Split(type=split, description=f'Unofficial speaker independent {split} split')

        split_df = samples_df[samples_df['actor'].isin(split_speakers)]
        # Emotion categories the actors tried to portray
        db[f'emotion.categories.desired.{split}'] = audformat.Table(
            index=split_df.index, split_id=split, description='The categorical emotions the actors were asked '
                                                              f' to portray (unofficial speaker independent '
                                                              f'{split} split).'
        )
        db[f'emotion.categories.desired.{split}']['emotion'] = audformat.Column(
            scheme_id='emotion',
        )
        db[f'emotion.categories.desired.{split}']['emotion'].set(split_df['emotion_expressed'])


        # Emotion categories the raters perceived

        db[f'emotion.categories.{split}.votes'] = audformat.Table(
            index=split_df.index, description='The votes for each emotion (unofficial speaker '
                                              f'independent {split} split)'
        )
        for emotion in emotion_labels:
            if emotion == 'no_agreement':
                continue
            db[f'emotion.categories.{split}.votes'][emotion] = audformat.Column(
                scheme_id='votes', description=f'The number of times raters voted for {emotion}.'
            )
            db[f'emotion.categories.{split}.votes'][emotion].set(
                votes_df.loc[split_df.index, emotion]
            )

        # Gold standard of emotion categories (by raters)

        db[f'emotion.categories.{split}.gold_standard'] = audformat.Table(
            index=split_df.index, description=f'The gold standard table of the unofficial speaker independent '
                                              f'{split} split for emotion categories.'
        )

        db[f'emotion.categories.{split}.gold_standard']['emotion'] = audformat.Column(
            scheme_id='emotion', description=f'The most commonly voted for categorical emotion, or no_agreement '
                                             f'when there is more than one.'
        )
        db[f'emotion.categories.{split}.gold_standard']['emotion'].set(
            split_df['emotion_recognized']
        )
        db[f'emotion.categories.{split}.gold_standard']['emotion.agreement'] = audformat.Column(
            scheme_id='emotion.agreement', description=f'The proportion of raters that voted for the most commonly '
                                                       f'selected emotion.'
        )
        db[f'emotion.categories.{split}.gold_standard']['emotion.agreement'].set(
            split_df['emotion.agreement']
        )

        if not os.path.exists(os.path.join(build_dir, audio_dir_name)):
            shutil.copytree(
                audio_path,
                os.path.join(current_dir, build_dir, audio_dir_name)
            )
        db.save(build_dir)

def gold_standard(votes):
    mode = pd.Series.mode(votes)
    if len(mode)==1:
        return mode
    else:
        return 'no_agreement'

def categorical_voter_percentage(votes):
    mode = pd.Series.mode(votes)
    if len(mode) == 0:
        return np.NaN
    mode = mode[0]
    max_n_votes = sum(votes == mode)
    total_votes = len(votes)
    return max_n_votes / total_votes

if __name__ == '__main__':
    main()



# export.py
# export the loaded training and test sets to it's own folder

import os
import pandas as pd
import configparser
import audeer
import argparse
import audiofile
from nkululeko.experiment import Experiment
from nkululeko.utils.util import Util
from nkululeko.constants import VERSION
import shutil


def main(src_dir):
    parser = argparse.ArgumentParser(description="Call the nkululeko framework.")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    util = Util("export")
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}")

    # export
    df_train = expr.df_train
    df_test = expr.df_test
    target_root = util.config_val("EXPORT", "root", "./exported_data/")
    orig_root = util.config_val("EXPORT", "orig_root", None)
    data_name = util.config_val("EXPORT", "data_name", "export")
    segments_as_files = eval(util.config_val("EXPORT", "segments_as_files", "False"))
    audeer.mkdir(target_root)
    splits = {"train": df_train, "test": df_test}
    df_all = pd.DataFrame()
    for split in splits:
        files = []
        df = splits[split]
        for idx, (file, start, end) in enumerate(df.index.to_list()):
            file_dir = os.path.dirname(file)
            if segments_as_files:
                signal, sampling_rate = audiofile.read(
                    file,
                    offset=start.total_seconds(),
                    duration=(end - start).total_seconds(),
                    always_2d=True,
                )
                file_name = (
                    os.path.splitext(file)[0] + "_" + start.total_seconds() + ".wav"
                )
                wav_folder = (
                    f"{target_root}/{os.path.basename(os.path.normpath(orig_root))}"
                )
                audeer.mkdir(wav_folder)
                new_rel_path = file_dir[
                    file_dir.index(orig_root) + 1 + len(orig_root) :
                ]
                new_file_path = f"{wav_folder}/{new_rel_path}"
                audeer.mkdir(new_file_path)
                new_file_name = f"{new_file_path}/{file_name}"
                audiofile.write(new_file_name, signal, sampling_rate)
                new_file_name = os.path.relpath(new_file_name, target_root)
                files.append(new_file_name)
            else:
                file_name = os.path.basename(file)
                wav_folder = (
                    f"{target_root}/{os.path.basename(os.path.normpath(orig_root))}"
                )
                audeer.mkdir(wav_folder)
                new_rel_path = file_dir[
                    file_dir.index(orig_root) + 1 + len(orig_root) :
                ]
                new_file_path = f"{wav_folder}/{new_rel_path}"
                audeer.mkdir(new_file_path)
                new_file_name = f"{new_file_path}/{file_name}"
                if not os.path.exists(new_file_name):
                    shutil.copyfile(file, new_file_name)
                new_file_name = os.path.relpath(new_file_name, target_root)
                files.append(new_file_name)
        df = df.set_index(df.index.set_levels(files, level="file"))
        df["split"] = split
        df_all = pd.concat([df_all, df])
    # remove encoded labels
    target = util.config_val("DATA", "target", "emotion")
    if "class_label" in df_all.columns:
        df_all = df_all.drop(columns=[target])
        df_all = df_all.rename(columns={"class_label": target})

    df_all.to_csv(f"{target_root}/{data_name}.csv")
    util.debug(f"saved {data_name}.csv to {target_root}, {df.shape[0]} samples.")

    print("DONE")


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # use this if you want to state the config file path on command line



""""
A predictor for age.
Currently based on audEERING's agender model.
"""
from nkululeko.utils.util import Util
from nkululeko.feature_extractor import FeatureExtractor
import ast
import nkululeko.glob_conf as glob_conf


class AgePredictor:
    """
    AgePredictor
    predicting age with the audEERING agender model

    """

    def __init__(self, df):
        self.df = df
        self.util = Util("agePredictor")

    def predict(self, split_selection):
        self.util.debug(f"predicting age for {split_selection} samples")
        feats_name = "_".join(ast.literal_eval(glob_conf.config["DATA"]["databases"]))
        self.feature_extractor = FeatureExtractor(
            self.df, ["agender_agender"], feats_name, split_selection
        )
        agender_df = self.feature_extractor.extract()
        pred_age = agender_df.age * 100
        #        pred_gender = agender_df.drop('age', axis=1).idxmax(axis=1)
        return_df = self.df.copy()
        #        return_df['gender_pred'] = pred_gender
        return_df["age_pred"] = pred_age.astype("int")
        return return_df



# process_databse.py --> SAVEE
# modified from https://github.com/Strong-AI-Lab/emotion/blob/master/datasets/SAVEE/process.py

import argparse
import re
from pathlib import Path

import pandas as pd

REGEX = re.compile(r"^([a-z][a-z]?)[0-9][0-9]$")

emotion_map = {
    "a": "anger",
    "d": "disgust",
    "f": "fear",
    "h": "happiness",
    "n": "neutral",
    "sa": "sadness",
    "su": "surprise",
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="SAVEE", help="Path of SAVEE directory (AudioData renamed to SAVEE)")
    parser.add_argument("--out_dir", type=str, default=".")
    args = parser.parse_args()
    
    data_dir = Path(args.data_dir)
    out_dir = Path(args.out_dir)

    wav_files = list(data_dir.glob("**/*.wav"))
    file = [wav_file for wav_file in wav_files]
    speaker = [wav_file.parts[1] for wav_file in wav_files]
    emotion = [emotion_map[REGEX.match(wav_file.stem).group(1)] for wav_file in wav_files if REGEX.match(wav_file.stem) is not None]
    gender = ["male" for wav_file in wav_files]
    language = ["english" for wav_file in wav_files]
    
    # convert to df
    df = pd.DataFrame({"file": file, "speaker": speaker, "emotion": emotion, 
        "language": language, "gender": gender})

    # allocate the last speaker, KL, for test set
    test_df = df[df["speaker"] == "KL"]
    train_df = df.drop(test_df.index)

    # save to CSV
    train_df.to_csv(out_dir / "savee_train.csv", index=False)
    test_df.to_csv(out_dir / "savee_test.csv", index=False)

    print(f"Total: {len(df)}, Train: {len(train_df)}, Test: {len(test_df)}")
if __name__ == "__main__":
    main()



# plots.py
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
import seaborn as sns
import numpy as np
import ast
from scipy import stats
from nkululeko.utils.util import Util
import nkululeko.utils.stats as su
import nkululeko.glob_conf as glob_conf
from nkululeko.reporting.report_item import ReportItem
from nkululeko.reporting.defines import Header


class Plots:
    def __init__(self):
        """Initializing the util system"""
        self.util = Util("plots")
        self.format = self.util.config_val("PLOT", "format", "png")
        self.target = self.util.config_val("DATA", "target", "emotion")

    def plot_distributions_speaker(self, df):
        df_speakers = pd.DataFrame()
        pd.options.mode.chained_assignment = None  # default='warn'
        for s in df.speaker.unique():
            df_speaker = df[df.speaker == s]
            df_speaker["samplenum"] = df_speaker.shape[0]
            df_speakers = pd.concat([df_speakers, df_speaker.head(1)])
        # plot the distribution of samples per speaker
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        self.util.debug(f"plotting samples per speaker")
        if "gender" in df_speakers:
            filename = f"samples_value_counts"
            ax = (
                df_speakers.groupby("samplenum")["gender"]
                .value_counts()
                .unstack()
                .plot(
                    kind="bar",
                    stacked=True,
                    title=f"samples per speaker ({df_speakers.shape[0]})",
                    rot=0,
                )
            )
            ax.set_ylabel(f"number of speakers")
            ax.set_xlabel("number of samples")
            self._save_plot(
                ax,
                "Samples per speaker",
                f"Samples per speaker ({df_speakers.shape[0]})",
                filename,
                "speakers",
            )

            # fig.clear()
        else:
            filename = f"samples_value_counts"
            ax = (
                df_speakers["samplenum"]
                .value_counts()
                .sort_values()
                .plot(
                    kind="bar",
                    stacked=True,
                    title=f"samples per speaker ({df_speakers.shape[0]})",
                    rot=0,
                )
            )
            ax.set_ylabel(f"number of speakers")
            ax.set_xlabel("number of samples")
            self._save_plot(
                ax,
                "Sample value counts",
                f"Samples per speaker ({df_speakers.shape[0]})",
                filename,
                "speakers",
            )

        self.plot_distributions(df_speakers, type_s="speakers")

    def plot_distributions(self, df, type_s="samples"):
        class_label, df = self._check_binning("class_label", df)
        attributes = ast.literal_eval(
            self.util.config_val("EXPL", "value_counts", False)
        )
        # always plot the distribution of the main attribute
        filename = f"{class_label}_distribution"
        if self.util.is_categorical(df[class_label]):
            ax = df[class_label].value_counts().plot(kind="bar")
        else:
            # for continuous variables, also add a discretized version
            binned_data = self.util.continuous_to_categorical(df[class_label])
            ax = binned_data.value_counts().plot(kind="bar")
            filename_binned = f"{class_label}_discreet"
            self._save_plot(
                ax,
                "Sample value counts",
                filename_binned,
                filename_binned,
                type_s,
            )
            dist_type = self.util.config_val("EXPL", "dist_type", "hist")
            ax = df[class_label].plot(kind=dist_type)

        self._save_plot(
            ax,
            "Sample value counts",
            filename,
            filename,
            type_s,
        )

        for att in attributes:
            if len(att) == 1:
                att1 = att[0]
                if att1 == self.target:
                    self.util.debug(f"no need to correlate {att1} with itself")
                    return
                if att1 not in df:
                    self.util.error(f"unknown feature: {att1}")
                att1, df = self._check_binning(att1, df)
                self.util.debug(f"plotting {att1}")
                filename = f"{self.target}-{att1}"
                if self.util.is_categorical(df[class_label]):
                    if self.util.is_categorical(df[att1]):
                        ax, caption = self._plot2cat(
                            df, class_label, att1, self.target, type_s
                        )
                    else:
                        ax, caption = self._plotcatcont(
                            df, class_label, att1, att1, type_s
                        )
                else:
                    if self.util.is_categorical(df[att1]):
                        ax, caption = self._plotcatcont(
                            df, att1, class_label, att1, type_s
                        )
                    else:
                        ax, caption = self._plot2cont(df, class_label, att1, type_s)
                self._save_plot(
                    ax,
                    caption,
                    f"Correlation of {self.target} and {att[0]}",
                    filename,
                    type_s,
                )
                # fig.clear()           # avoid error
            elif len(att) == 2:
                att1 = att[0]
                att2 = att[1]
                if att1 == self.target or att2 == self.target:
                    self.util.debug(f"no need to correlate {self.target} with itself")
                    return
                if att1 not in df:
                    self.util.error(f"unknown feature: {att1}")
                if att2 not in df:
                    self.util.error(f"unknown feature: {att2}")
                att1, df = self._check_binning(att1, df)
                att2, df = self._check_binning(att2, df)
                self.util.debug(f"plotting {att}")
                filename = f"{att1}-{att2}"
                filename = f"{self.target}-{filename}"
                if self.util.is_categorical(df["class_label"]):
                    if self.util.is_categorical(df[att1]):
                        if self.util.is_categorical(df[att2]):
                            # class_label = cat, att1 = cat, att2 = cat
                            ax, caption = self._plot2cat(df, att1, att2, att1, type_s)
                        else:
                            # class_label = cat, att1 = cat, att2 = cont
                            ax, caption = self._plotcatcont(
                                df, att1, att2, att1, type_s
                            )
                    else:
                        if self.util.is_categorical(df[att2]):
                            # class_label = cat, att1 = cont, att2 = cat
                            ax, caption = self._plotcatcont(
                                df, att2, att1, att2, type_s
                            )
                        else:
                            # class_label = cat, att1 = cont, att2 = cont
                            ax, caption = self._plot2cont_cat(
                                df, att1, att2, "class_label", type_s
                            )
                else:  # class_label is continuous
                    if self.util.is_categorical(df[att1]):
                        if self.util.is_categorical(df[att2]):
                            # class_label = cont, att1 = cat, att2 = cat
                            ax, caption = self._plot2cat(df, att1, att2, att1, type_s)
                        else:
                            # class_label = cont, att1 = cat, att2 = cont
                            ax, caption = self._plot2cont_cat(
                                df, att2, "class_label", att1, type_s
                            )
                    else:
                        if self.util.is_categorical(df[att2]):
                            # class_label = cont, att1 = cont, att2 = cat
                            ax, caption = self._plot2cont_cat(
                                df, att1, "class_label", att2, type_s
                            )
                        else:
                            # class_label = cont, att1 = cont, att2 = cont
                            ax, caption = self._plot2cont(df, att1, att2, type_s)

                self._save_plot(
                    ax, caption, f"Correlation of {att1} and {att2}", filename, type_s
                )

            else:
                self.util.error(
                    "plot value counts: the plot distribution descriptor for"
                    f" {att} has more than 2 values. Perhaps you forgot to state a list of lists?"
                )

    def _save_plot(self, ax, caption, header, filename, type_s):
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        fig = ax.figure
        # avoid warning
        # plt.tight_layout()
        img_path = f"{fig_dir}{filename}_{type_s}.{self.format}"
        plt.savefig(img_path)
        plt.close(fig)
        # fig.clear()   # avoid error
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                header,
                caption,
                img_path,
            )
        )

    def _check_binning(self, att, df):
        bin_reals_att = eval(self.util.config_val("EXPL", f"{att}.bin_reals", "False"))
        if bin_reals_att:
            self.util.debug(f"binning continuous variable {att} to categories")
            att_new = f"{att}_binned"
            df[att_new] = self.util.continuous_to_categorical(df[att]).values
            att = att_new
        return att, df

    def _plot2cont_cat(self, df, cont1, cont2, cat, ylab):
        """
        plot relation of two continuous distributions with one categorical
        """
        pearson = stats.pearsonr(df[cont1], df[cont2])
        # trunc to three digits
        pearson = int(pearson[0] * 1000) / 1000
        pearson_string = f"PCC: {pearson}"
        ax = sns.lmplot(data=df, x=cont1, y=cont2, hue=cat)
        caption = f"{ylab} {df.shape[0]}. {pearson_string}"
        ax.fig.suptitle(caption)
        return ax, caption

    def _plot2cont(self, df, col1, col2, ylab):
        """
        plot relation of two continuous distributions
        """
        pearson = stats.pearsonr(df[col1], df[col2])
        # trunc to three digits
        pearson = int(pearson[0] * 1000) / 1000
        pearson_string = f"PCC: {pearson}"
        ax = sns.lmplot(data=df, x=col1, y=col2)
        caption = f"{ylab} {df.shape[0]}. {pearson_string}"
        ax.fig.suptitle(caption)
        return ax, caption

    def _plotcatcont(self, df, cat_col, cont_col, xlab, ylab):
        """
        plot relation of categorical distribution with continuous
        """
        dist_type = self.util.config_val("EXPL", "dist_type", "hist")
        cats, cat_str, es = su.get_effect_size(df, cat_col, cont_col)
        if dist_type == "hist":
            ax = sns.histplot(df, x=cont_col, hue=cat_col, kde=True)
            caption = f"{ylab} {df.shape[0]}. {cat_str} ({cats}):" f" {es}"
            ax.set_title(caption)
            ax.set_xlabel(f"{cont_col}")
            ax.set_ylabel(f"number of {ylab}")
        else:
            ax = sns.displot(
                df, x=cont_col, hue=cat_col, kind="kde", fill=True, warn_singular=False
            )
            ax.set(xlabel=f"{cont_col}")
            caption = f"{ylab} {df.shape[0]}. {cat_str} ({cats}):" f" {es}"
            ax.fig.suptitle(caption)
        return ax, caption

    def _plot2cat(self, df, col1, col2, xlab, ylab):
        """
        plot relation of 2 categorical distributions
        """
        crosstab = pd.crosstab(index=df[col1], columns=df[col2])
        res_pval = stats.chi2_contingency(crosstab)
        res_pval = int(res_pval[1] * 1000) / 1000
        caption = f"{ylab} {df.shape[0]}. P-val chi2: {res_pval}"
        ax = (
            df.groupby(col1, observed=False)[col2]
            .value_counts()
            .unstack()
            .plot(kind="bar", stacked=True, title=caption, rot=0)
        )
        ax.set_ylabel(f"number of {ylab}")
        ax.set_xlabel(xlab)
        return ax, caption

    def plot_durations(self, df, filename, sample_selection, caption=""):
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        try:
            ax = sns.histplot(df, x="duration", hue="class_label", kde=True)
        except AttributeError as ae:
            self.util.warn(ae)
            ax = sns.histplot(df, x="duration", kde=True)
        min = self.util.to_3_digits(df.duration.min())
        max = self.util.to_3_digits(df.duration.max())
        title = f"Duration distr. for {sample_selection} {df.shape[0]}. min={min}, max={max}"
        ax.set_title(title)
        ax.set_xlabel(f"duration")
        ax.set_ylabel(f"number of samples")
        fig = ax.figure
        # plt.tight_layout()
        img_path = f"{fig_dir}{filename}_{sample_selection}.{self.format}"
        plt.savefig(img_path)
        plt.close(fig)
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                caption,
                title,
                img_path,
            )
        )

    def describe_df(self, name, df, target, filename):
        """Make a stacked barplot of samples and speakers per sex and target values. speaker, gender and target columns must be present"""
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        sampl_num = df.shape[0]
        sex_col = "gender"
        if target == "gender":
            sex_col = "class_label"
        if self.util.exp_is_classification() and target != "gender":
            target = "class_label"
        if df.is_labeled:
            if df.got_gender and df.got_speaker:
                spkr_num = df.speaker.nunique()
                female_smpl_num = df[df[sex_col] == "female"].shape[0]
                male_smpl_num = df[df[sex_col] == "male"].shape[0]
                self.util.debug(
                    f"plotting {name}: # samples: {sampl_num} (f:"
                    f" {female_smpl_num}, m: "
                    + f"{male_smpl_num}), # speakers: {spkr_num}"
                )
                # fig, axes = plt.subplots(nrows=1, ncols=2)
                fig, axes = plt.subplots(nrows=1, ncols=1)
                # df.groupby(target)['gender'].value_counts().unstack().plot(kind='bar', stacked=True, ax=axes[0], \
                #     title=f'samples ({sampl_num})')
                df.groupby(target)["gender"].value_counts().unstack().plot(
                    kind="bar", stacked=True, title=f"samples ({sampl_num})"
                )
                # df.groupby(target)['speaker'].nunique().plot(kind='bar', ax=axes[1], title=f'speakers ({spkr_num})')
            else:
                self.util.debug(f"plotting {name}: # samples: {sampl_num}")
                fig, axes = plt.subplots(nrows=1, ncols=1)
                df[target].value_counts().plot(
                    kind="bar", ax=axes, title=f"samples ({sampl_num})"
                )
            # plt.tight_layout()
            img_path = f"{fig_dir}{filename}.{self.format}"
            plt.savefig(img_path)
            fig.clear()
            plt.close(fig)
            glob_conf.report.add_item(
                ReportItem(
                    Header.HEADER_EXPLORE,
                    f"Overview on {df.shape[0]} samples",
                    "",
                    img_path,
                )
            )

    def scatter_plot(self, feats, label_df, label, dimred_type):
        dim_num = int(self.util.config_val("EXPL", "scatter.dim", 2))
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        sample_selection = self.util.config_val("EXPL", "sample_selection", "all")
        filename = f"{label}_{self.util.get_feattype_name()}_{sample_selection}_{dimred_type}_{str(dim_num)}d"
        filename = f"{fig_dir}{filename}.{self.format}"
        self.util.debug(f"computing {dimred_type}, this might take a while...")
        data = None
        labels = label_df[label]
        if dimred_type == "tsne":
            data = self.getTsne(feats, dim_num)
        else:
            if dimred_type == "umap":
                import umap

                y = umap.UMAP(
                    n_neighbors=10,
                    random_state=0,
                    n_components=dim_num,
                ).fit_transform(feats.values)
            elif dimred_type == "pca":
                from sklearn.decomposition import PCA
                from sklearn.preprocessing import StandardScaler

                scaler = StandardScaler()
                pca = PCA(n_components=dim_num)
                y = pca.fit_transform(scaler.fit_transform(feats.values))
            else:
                self.util.error(
                    f"no such dimensionality reduction function: {dimred_type}"
                )
            if dim_num == 2:
                columns = ["Dim_1", "Dim_2"]
            elif dim_num == 3:
                columns = ["Dim_1", "Dim_2", "Dim_3"]
            else:
                self.util.error(f"wrong dimension number: {dim_num}")
            data = pd.DataFrame(
                y,
                feats.index,
                columns=columns,
            )

        if dim_num == 2:
            plot_data = np.vstack((data.T, labels)).T
            plot_df = pd.DataFrame(data=plot_data, columns=("Dim_1", "Dim_2", "label"))
            # plt.tight_layout()
            ax = (
                sns.FacetGrid(plot_df, hue="label", height=6)
                .map(plt.scatter, "Dim_1", "Dim_2")
                .add_legend()
            )
        elif dim_num == 3:
            from mpl_toolkits.mplot3d import Axes3D
            from sklearn.preprocessing import LabelEncoder

            le = LabelEncoder()

            labels_e = le.fit_transform(labels)
            plot_data = np.vstack((data.T, labels_e)).T
            plot_df = pd.DataFrame(
                data=plot_data, columns=("Dim_1", "Dim_2", "Dim_3", "label")
            )
            # plt.tight_layout()
            # axes instance
            fig = plt.figure(figsize=(6, 6))
            ax = Axes3D(fig, auto_add_to_figure=False)
            fig.add_axes(ax)
            # get colormap from seaborn
            # cmap = ListedColormap(sns.color_palette("hsv", 256).as_hex())
            color_dict = {
                0: "red",
                1: "blue",
                2: "green",
                3: "yellow",
                4: "purple",
                5: "#ff69b4",
                6: "black",
                7: "cyan",
                8: "magenta",
                9: "#faebd7",
                10: "#2e8b57",
                11: "#eeefff",
                12: "#da70d6",
                13: "#ff7f50",
                14: "#cd853f",
                15: "#bc8f8f",
                16: "#5f9ea0",
                17: "#daa520",
            }
            # plot
            # make the numbers bigger so they can be used as distinguishable colors
            labels_ex = [color_dict[xi] for xi in labels_e]
            sc = ax.scatter(
                plot_df.Dim_1,
                plot_df.Dim_2,
                plot_df.Dim_3,
                s=40,
                c=labels_ex,
                marker="o",
                # cmap=cmap,
                alpha=1,
            )
            ax.set_xlabel("Dim_1")
            ax.set_ylabel("Dim_2")
            ax.set_zlabel("Dim_3")
            # legend
            plt.legend(*sc.legend_elements(), bbox_to_anchor=(1.05, 1), loc=2)
        else:
            self.util.error(f"wrong dimension number: {dim_num}")
        fig = ax.figure
        plt.savefig(filename)
        fig.clear()
        plt.close(fig)
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                f"Scatter plot",
                f"using {dimred_type}",
                filename,
            )
        )

    def getTsne(self, feats, dim_num, perplexity=30, learning_rate=200):
        """Make a TSNE plot to see whether features are useful for classification"""
        model = TSNE(
            n_components=dim_num,
            random_state=0,
            perplexity=perplexity,
            learning_rate=learning_rate,
        )
        tsne_data = model.fit_transform(feats)
        return tsne_data

    def plot_feature(self, title, feature, label, df_labels, df_features):
        # remove fullstops in the name
        feature_name = feature.replace(".", "-")
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        filename = f"{fig_dir}feat_dist_{title}_{feature_name}.{self.format}"
        if self.util.is_categorical(df_labels[label]):
            df_plot = pd.DataFrame(
                {label: df_labels[label], feature: df_features[feature]}
            )
            ax = sns.violinplot(data=df_plot, x=label, y=feature)
            label = self.util.config_val("DATA", "target", "class_label")
            ax.set(title=f"{title} samples", xlabel=label)
        else:
            plot_df = pd.concat([df_labels, df_features], axis=1)
            ax, caption = self._plot2cont(plot_df, label, feature, feature)
        # def _plot2cont(self, df, col1, col2, xlab, ylab):

        fig = ax.figure
        # plt.tight_layout()
        plt.savefig(filename)
        fig.clear()
        plt.close(fig)
        caption = f"Feature plot for feature {feature}"
        content = caption
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                caption,
                content,
                filename,
            )
        )

    def plot_tree(self, model, features):
        from sklearn import tree

        ax = plt.gca()
        ax.figure.set_size_inches(100, 60)
        #        tree.plot_tree(model, ax = ax)
        tree.plot_tree(model, feature_names=list(features.columns), ax=ax)
        # plt.tight_layout()
        # print(ax)
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        exp_name = self.util.get_exp_name(only_data=True)
        format = self.util.config_val("PLOT", "format", "png")
        filename = f"{fig_dir}{exp_name}EXPL_tree-plot.{format}"
        fig = ax.figure
        fig.savefig(filename)
        fig.clear()
        plt.close(fig)
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                f"Tree plot",
                f"for feature importance",
                filename,
            )
        )



# model_bayes.py

from sklearn.naive_bayes import GaussianNB
from nkululeko.models.model import Model


class Bayes_model(Model):
    is_classifier = True

    """A Bayesian model"""

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.clf = GaussianNB()  # set up the classifier



# process_database.py
# process IEMOCAP database to match the format of Nkululeko
# adopted from S3PRL emotion downstream task

import argparse
import glob
import os
import re
from os.path import basename
from os.path import join as path_join
from os.path import splitext

import pandas as pd

LABEL_DIR_PATH = 'dialog/EmoEvaluation'
WAV_DIR_PATH = 'sentences/wav'


def get_wav_paths(data_dirs):
    wav_paths = glob.glob(f"{data_dirs}/**/*.wav", recursive=True)
    wav_dict = {}
    for wav_path in wav_paths:
        wav_name = splitext(basename(wav_path))[0]
        start = wav_path.find('Session')
        wav_path = wav_path[start:]
        wav_dict[wav_name] = wav_path

    return wav_dict


def preprocess(data_dirs, paths, out_path):
    meta_data = []
    for path in paths:
        wav_paths = get_wav_paths(path_join(data_dirs, path, WAV_DIR_PATH))
        label_dir = path_join(data_dirs, path, LABEL_DIR_PATH)
        label_paths = list(os.listdir(label_dir))
        label_paths = [label_path for label_path in label_paths
                       if splitext(label_path)[1] == '.txt']
        for label_path in label_paths:
            with open(path_join(label_dir, label_path)) as f:
                for line in f:
                    if line[0] != '[':
                        continue
                    line = re.split('[\t\n]', line)
                    line = list(filter(None, line))
                    if line[2] not in ['neu', 'hap', 'ang', 'sad', 'exc']:
                        continue
                    if line[1] not in wav_paths:
                        continue
                    meta_data.append({
                        'file': os.path.join(data_dirs, wav_paths[line[1]]),
                        'emotion': line[2].replace('exc', 'hap'),
                        'speaker': re.split('_', basename(wav_paths[line[1]]))[0],
                        'gender': re.split('_', basename(wav_paths[line[1]]))[-1][0]
                    })
    data = {
        'labels': {'neu': 0, 'hap': 1, 'ang': 2, 'sad': 3},
        'meta_data': meta_data
    }

    # print number of utterances
    print(f"Total number of {os.path.basename(out_path)}: {len(data['meta_data'])}")

    # save to CSV file
    pd.DataFrame(data['meta_data']).to_csv(out_path, index=False)


def main(data_dir):
    """Main function."""
    paths = list(os.listdir(data_dir))
    paths = [path for path in paths if path[:7] == 'Session']
    paths.sort()
    out_dir = os.path.dirname(__file__)
    # make Session 1-3 as training set, Sesion4 for dev, Session 5 as test set
    preprocess(data_dir, ['Session5'], path_join(out_dir, 'iemocap_test.csv'))
    preprocess(data_dir, ['Session4'], path_join(out_dir, 'iemocap_dev.csv'))
    preprocess(data_dir, ['Session1', 'Session2', 'Session3'], path_join(out_dir, 'iemocap_train.csv'))

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='./IEMOCAP_full_release')
    args = parser.parse_args()
    main(args.data_dir)



# demo.py
# Demonstration code to use the ML-experiment framework
# Test the loading of a previously trained model and demo mode
# needs the project config file to run before

from nkululeko.experiment import Experiment
import configparser
from nkululeko.utils.util import Util
from nkululeko.constants import VERSION
import argparse
import os


def main(src_dir):
    parser = argparse.ArgumentParser(description="Call the nkululeko framework.")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    parser.add_argument(
        "--file", help="A file that should be processed (16kHz mono wav)"
    )
    parser.add_argument(
        "--list",
        help=(
            "A file with a list of files, one per line, that should be"
            " processed (16kHz mono wav)"
        ),
    )
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)

    # create a new experiment
    expr = Experiment(config)
    util = Util("demo")
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    # load the experiment
    expr.load(f"{util.get_save_name()}")
    if args.file is None and args.list is None:
        expr.demo(None, False)
    else:
        if args.list is None:
            expr.demo(args.file, False)
        else:
            expr.demo(args.list, True)

    print("DONE")


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # use this if you want to state the config file path on command line



# main.py
# Demonstration code to use the ML-experiment framework

import sys
sys.path.append("./src")
from reporter import Reporter
import experiment as exp
import configparser
from util import Util

def main(config_file):
    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    util = Util()
    
    # create a new experiment
    expr = exp.Experiment(config)
    print(f'running {expr.name}')

    # load the experiment
    expr.load(f'{util.get_exp_name()}.pkl')

    expr.demo()

    print('DONE')


if __name__ == "__main__":
    main(sys.argv[1])



#!/usr/bin/env python3
# process_database.py --> DEMoS

import argparse
from pathlib import Path

import pandas as pd

emotion_map = {
    "rab": "anger",
    "tri": "sadness",
    "gio": "happiness",
    "pau": "fear",
    "dis": "disgust",
    "col": "guilt",
    "sor": "surprise",
    "neu": "neutral",
}

gender_map = {
    "f": "female",
    "m": "male"
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='DEMoS', help='Path to the database file')
    parser.add_argument('--output_dir', type=str, default='.', help='Path to the output directory')
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    if not output_dir.exists():
        output_dir.mkdir()

    paths = list(data_dir.glob('**/*.wav'))
    # emotions = [emotion_map[p.stem.split('_')[3][:3]] if p.parts[1] == 'DEMOS' else emotion_map[p.stem.split('_')[2][:3]] for p in paths]

    emotions = [emotion_map[p.stem.split('_')[-1][:3]] for p in paths]
    genders = [gender_map[p.stem.split('_')[-3]] for p in paths]
    speakers = [p.stem.split('_')[-2] for p in paths]
    languages = ["italian" for p in paths]
    prototypicality = [p.stem[:2] if p.stem[:2] in {"NP", "PR"} else "neutral" for p in paths]


    # convert to df
    df = pd.DataFrame({"file": paths, "emotion": emotions, "gender": genders, "speaker": speakers, "language": languages, "prototypicality": prototypicality})

    # split train and test based in speaker independent and balanced emotion
    # allocate speakers >= 55 for test
    df_test = df[df["speaker"] > "57"]  #python3.9 newer 
    # df_test = df[df["speaker"].astype(int) > 57]
    df_train = df.drop(df_test.index)

    # save to csv
    df.to_csv(output_dir / "demos.csv", index=False)
    df_train.to_csv(output_dir / "demos_train.csv", index=False)
    df_test.to_csv(output_dir / "demos_test.csv", index=False)

    print(f"Total: {len(df)}, Train: {len(df_train)}, Test: {len(df_test)}")


if __name__ == "__main__":
    main()

